/* ... */
/* this file is part of netrafd project */
/* Copyright (c) 2005 M.S. */


#include "menus.h"
#include "gui.h"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>




  /**
    Thread "semaphore" allowing only one refresh-screen request
    to be done at one moment.
  */
  extern pthread_mutex_t refresh_locker;




/**
  Looking through table of strings (NULL terminated)
  and returns length of longest string.
  @param strtable  Table of strings (NULL terminated)
  @return  length of longest string
*/
size_t longest_strlen(char **strtable){
  int table_size;
  size_t longest=0, current;
  int i;


  /* count table size */
  NULL_TERMINATED_ARRAY_SIZE(table_size, strtable);
  
  /* find length of longest string */
  for(i=0; i<table_size-1; i++){
    current = strlen(strtable[i]);
    if(longest<current)
      longest = current;
  }

  return longest;
}




/**
  Create 1-collumn menu.
  @param begin_y  Upper-left corner start-line
  @param begin_x  Upper-left corner start-collumn
  @param fore_attr  Foreground (selected item) ncurses color_pair
  @param back_attr  Background (everything besides selected item) ncurses color-pair
  @param choices  Pointer to string table of menu choices (NULL-terminated)
  @param descriptions  Pointer to string table of menu descriptions (NULL-terminated)
  @param visible_rows  If visible_rows<=0 then height of menu will be counted automatially,
                       otherwise this parameter describes maximum number of visible rows.
		       If number of menu items is greater than visible_rows parameter,
		       menu will be scrollable.
  @param forced_width  When forced_width>0, menu will have exact "forced_width" characters width
  @param winname  Frame around can be named with "winname" string. If you don't want
                  named frame pass NULL.
  @return  Pointer to allocated and initialized MENU_T structure or NULL if fail
*/
MENU_T *create_sub_menu(int begin_y,
		        int begin_x,
			chtype fore_attr,
			chtype back_attr,
			char **choices,
			char **descriptions,
			int visible_rows,
			int forced_width,
			char *winname){
  MENU_T *menu;
  int nr_of_choices;
  size_t menu_width;
  int i;


  /* count menu width */
  if(forced_width > 0)
    menu_width = (size_t)forced_width + 2;
  else
    menu_width = longest_strlen(choices) + 2;

  /* allocate heap memory to storage MENU_T structure */
  menu = (MENU_T*)malloc(sizeof(MENU_T));
  if(menu == NULL)
    return NULL;
  NULL_TERMINATED_ARRAY_SIZE(nr_of_choices, choices);
  
  /* create menu items */
  menu->menu_items = (ITEM**)calloc(nr_of_choices, sizeof(ITEM*));
  for(i=0; i<nr_of_choices; i++)
    menu->menu_items[i] = new_item(choices[i], descriptions[i]);

  /* create menu structure */
  menu->menu = new_menu(menu->menu_items);

  /* set menu options */
  menu_opts_off(menu->menu, O_SHOWDESC | O_NONCYCLIC);
  if(visible_rows > 0)
    set_menu_format(menu->menu, visible_rows, 1);
  else
    set_menu_format(menu->menu, nr_of_choices-1, 1);
  set_menu_mark(menu->menu, NULL);
  if(forced_width > 0){
    menu->menu->width = forced_width;
    menu->menu->namelen = forced_width;
  }
  
  /* create the window and panel to be associated with menu */
  if(visible_rows > 0)
    menu->menu_win = newwin(visible_rows+2, menu_width, begin_y, begin_x);
  else
    menu->menu_win = newwin((nr_of_choices-1)+2, menu_width, begin_y, begin_x);
  keypad(menu->menu_win, TRUE);
  menu->menu_panel = new_panel(menu->menu_win);
  set_menu_win(menu->menu, menu->menu_win);
  if(visible_rows > 0)
    set_menu_sub(menu->menu, derwin(menu->menu_win, visible_rows, menu_width-2, 1, 1));
  else
    set_menu_sub(menu->menu, derwin(menu->menu_win, nr_of_choices-1, menu_width-2, 1, 1));
  wbkgd(menu_sub(menu->menu), back_attr);

  /* draw box around menu */
  wattron(menu->menu_win, back_attr);
  box(menu->menu_win, 0, 0);
  if(winname != NULL)
    mvwprintw(menu->menu_win, 0, 2, "%s", winname);
  wattroff(menu->menu_win, back_attr);

  /* set menu colors */
  set_menu_fore(menu->menu, fore_attr);
  set_menu_back(menu->menu, back_attr);

  /* post menu */
  post_menu(menu->menu);
  hide_panel(menu->menu_panel);

  /* sub menu doesn't have any sub-menus... */
  menu->sub_menus = NULL;
  
  /* sub menu doesn't have status window or panel too... */
  menu->status_win = NULL;
  menu->status_panel = NULL;
  
  return menu;
}




/**
  Create main menu (bar) and all submenus.
  Beside first 4 arguments function takes as many arguments as main_choices table size.
  All of that arguments are in form:
    char **sub_menu_1_choices,  char **sub_menu_1_descriptions,
    char **sub_menu_2_choices,  char **sub_menu_2_descriptions,
                   ...
		   ...
    char **sub_menu_N_choices,  char **sub_menu_N_descriptions,
  @param fore_attr  Foreground (selected item) ncurses color_pair
  @param back_attr  Background (everything besides selected item) ncurses color-pair
  @param main_choices  Pointer to string table of main_menu choices (NULL-terminated)
  @param main_descriptions  Pointer to string table of main_menu descriptions (NULL-terminated)
  @return  Pointer to allocated and initialized MENU_T structure or NULL if fail
*/
MENU_T *create_menu(chtype fore_attr,
		    chtype back_attr,
		    char **main_choices,
		    char **main_descriptions,
		    ...){
  MENU_T *main_menu;
  int nr_of_main_choices;
  va_list ap;
  char **choices, **descriptions;
  int i, x=0;


  /* initialize variable argument list */
  va_start(ap, main_descriptions);
  
  /* allocate heap memory to storage MENU_T structure */
  main_menu = (MENU_T*)calloc(1, sizeof(MENU_T));
  if(main_menu == NULL)
    return NULL;
  NULL_TERMINATED_ARRAY_SIZE(nr_of_main_choices, main_choices);

  /* create main menu items */
  main_menu->menu_items = (ITEM**)calloc(nr_of_main_choices, sizeof(ITEM*));
  for(i=0; i<nr_of_main_choices; i++)
    main_menu->menu_items[i] = new_item(main_choices[i], main_descriptions[i]);

  /* create main menu structure */
  main_menu->menu = new_menu(main_menu->menu_items);

  /* set menu options */
  menu_opts_off(main_menu->menu, O_SHOWDESC | O_NONCYCLIC);
  set_menu_format(main_menu->menu, 1, nr_of_main_choices-1);
  set_menu_mark(main_menu->menu, NULL);
  
  /* create the window and panel to be associated with menu */
  main_menu->menu_win = newwin(1, COLS, 1, 0);
  draw_bar(main_menu->menu_win, "", 0, ALIGN_LEFT, back_attr);
  keypad(main_menu->menu_win, TRUE);
  main_menu->menu_panel = new_panel(main_menu->menu_win);
  set_menu_win(main_menu->menu, main_menu->menu_win);

  /* set menu colors */
  set_menu_fore(main_menu->menu, fore_attr);
  set_menu_back(main_menu->menu, back_attr);

  /* post menu */
  post_menu(main_menu->menu);
  hide_panel(main_menu->menu_panel);

  /* create submenus */
  main_menu->sub_menus = (MENU_T**)calloc(nr_of_main_choices, sizeof(MENU_T*));
  for(i=0; i<nr_of_main_choices-1; i++){
    choices = va_arg(ap, char**);
    descriptions = va_arg(ap, char**);
    main_menu->sub_menus[i] = create_sub_menu(2, x, 
                                              fore_attr, back_attr,
					      choices, descriptions, 0, 0, NULL);
    if(main_menu->sub_menus[i] == NULL){
      free(main_menu);
      return NULL;
    }
    x += strlen(main_choices[i]) + 1;
  }
  
  /* create window and panel for status-bar help (descriptions) */
  main_menu->status_win = newwin(1, COLS, LINES-1, 0);
  keypad(main_menu->status_win, TRUE);
  main_menu->status_panel = new_panel(main_menu->status_win);
  hide_panel(main_menu->status_panel);

  /* clean variable argument list structures */
  va_end(ap);

  return main_menu;
}




/**
  Prints description of currently selected menu-item.
  @param menu  Pointer to MENU_T structure
               (it must be initialized with create_menu() function
*/
__inline__ void print_current_description(MENU_T *menu){


  draw_bar(menu->status_win,
           item_description(current_item(menu->sub_menus[item_index(current_item(menu->menu))]->menu)),
           0,
	   ALIGN_LEFT,
	   menu_back(menu->menu));

  return;
}




/**
  Command-processing loop for main menu system.
  For request and return values descriptions see menu_driver(3X).
  @param menu  Pointer to MENU_T structure
               (it must be initialized with create_menu() function
  @param request  Requests are as follows:
		  REQ_LEFT_ITEM, REQ_RIGHT_ITEM, REQ_UP_ITEM, REQ_DOWN_ITEM
  @return  E_OK if routine succeded
*/
int main_menu_driver(MENU_T *menu, int request){
  int answer;


  /* enter "critical section" */
  pthread_mutex_lock(&refresh_locker);

  /* process request and send them to apropriate menus */
  if(request==REQ_RIGHT_ITEM || request==REQ_LEFT_ITEM){
    hide_panel(menu->sub_menus[item_index(current_item(menu->menu))]->menu_panel);
    answer = menu_driver(menu->menu, request);
    show_panel(menu->sub_menus[item_index(current_item(menu->menu))]->menu_panel);
  }else if(request==REQ_DOWN_ITEM || request==REQ_UP_ITEM){
    answer = menu_driver(menu->sub_menus[item_index(current_item(menu->menu))]->menu, request);
  }else
    return E_UNKNOWN_COMMAND;
  print_current_description(menu);

  /* leave "critical section" */
  pthread_mutex_unlock(&refresh_locker);

  update_screen();

  return answer;
}




/**
  Process user input on menu and returns user's choice.
  For return values descriptions see menu_driver(3X).
  @param menu  Pointer to MENU_T structure
               (it must be initialized with create_menu() function
  @param main  It is RETURN value - pointer to selected main ITEM structure
  @param main  It is RETURN value - pointer to selected sub ITEM structure
  @return  E_OK if routine succeded
*/
int process_main_menu(MENU_T *menu, ITEM **main, ITEM **sub){
  int answer=E_OK;
  int ch;


  /* show apropriate panels and (main, sub and status window) */
  pthread_mutex_lock(&refresh_locker);
  show_panel(menu->menu_panel);
  show_panel(menu->sub_menus[item_index(current_item(menu->menu))]->menu_panel);
  show_panel(menu->status_panel);
  pthread_mutex_unlock(&refresh_locker);

  print_current_description(menu);
  update_screen();

  /* process keyboard input */
  while((ch = wgetch(menu->menu_win)) != 27){   /* #27 - ESC */
    switch(ch){
      case KEY_RIGHT:
	answer = main_menu_driver(menu, REQ_RIGHT_ITEM);
	break;
      case KEY_LEFT:
	answer = main_menu_driver(menu, REQ_LEFT_ITEM);
	break;
      case KEY_DOWN:
        answer = main_menu_driver(menu, REQ_DOWN_ITEM);
	break;
      case KEY_UP:
        answer = main_menu_driver(menu, REQ_UP_ITEM);
	break;
      case 10:   /* #10 - ENTER */
        pthread_mutex_lock(&refresh_locker);
        hide_panel(menu->menu_panel);
        hide_panel(menu->sub_menus[item_index(current_item(menu->menu))]->menu_panel);
        hide_panel(menu->status_panel);
        pthread_mutex_unlock(&refresh_locker);
	*main = current_item(menu->menu);
	*sub = current_item(menu->sub_menus[item_index(current_item(menu->menu))]->menu);
        return answer;
	break;
      default:
        break;
    }
  }
  
  /* when done - hide menu panels */
  pthread_mutex_lock(&refresh_locker);
  hide_panel(menu->menu_panel);
  hide_panel(menu->sub_menus[item_index(current_item(menu->menu))]->menu_panel);
  hide_panel(menu->status_panel);
  pthread_mutex_unlock(&refresh_locker);
  *main = NULL;
  *sub = NULL;

  return E_NO_MATCH;
}




/** 
  Frees all memory and clean data structures occupied by sub_menu.
  @param menu  Pointer to MENU_T structure
               (it must be initialized with create_sub_menu() function
*/
void destroy_sub_menu(MENU_T *menu){
  int nr_of_choices, i;
  WINDOW *sub_win;


  /* "remember" nr of items and pointer to sub_window */
  nr_of_choices = item_count(menu->menu);
  sub_win = menu_sub(menu->menu);

  pthread_mutex_lock(&refresh_locker);

  /* unpost and free MENU structure */
  if(free_menu(menu->menu) == E_POSTED){
    unpost_menu(menu->menu);
    free_menu(menu->menu);
  }

  /* delete menu panel, sub_window and window */
  del_panel(menu->menu_panel);
  delwin(sub_win);
  delwin(menu->menu_win);
  
  pthread_mutex_unlock(&refresh_locker);

  /* free all menu items and free menu_items array */
  for(i=0; i<nr_of_choices; i++)
    free_item(menu->menu_items[i]);
  free(menu->menu_items);

  /* finally - free MENU_T structure memory */
  free(menu);

  return;
}




/**
  Frees all memory and clean data structures occupied by menu.
  @param menu  Pointer to MENU_T structure
               (it must be initialized with create_menu() function
*/
void destroy_menu(MENU_T *menu){
  int nr_of_choices, i;


  /* "remember" nr of items */  
  nr_of_choices = item_count(menu->menu);
  
  pthread_mutex_lock(&refresh_locker);

  /* unpost and free MENU structure */
  if(free_menu(menu->menu) == E_POSTED){
    unpost_menu(menu->menu);
    free_menu(menu->menu);
  }

  /* delete status panel and window (if such one exist) */
  if(menu->status_panel != NULL)
    del_panel(menu->status_panel);
  if(menu->status_win != NULL)
    delwin(menu->status_win);

  /* delete menu panel and window */
  del_panel(menu->menu_panel);
  delwin(menu->menu_win);

  pthread_mutex_unlock(&refresh_locker);
  
  /* free all menu items and free menu_items array */
  for(i=0; i<nr_of_choices; i++)
    free_item(menu->menu_items[i]);
  free(menu->menu_items);

  /* clean and free all submenus */
  for(i=0; i<nr_of_choices; i++)
    destroy_sub_menu(menu->sub_menus[i]);
  free(menu->sub_menus);

  /* finally - free MENU_T structure memory */
  free(menu);

  return;
}
